import pandas as pd
import sortednp as snp

from collections import defaultdict, deque
from sortedcontainers import SortedSet
from math import inf
from heapq import heappop, heappush
from numpy import array

from realkd.logic import Conjunction, Constraint, KeyValueProposition, TabulatedProposition


class Node:
    """
    Represents a potential node (and incoming edge) for searches in the concept graph
    with edges representing the direct prefix-preserving successor relation (dpps).
    """

    def __init__(self, gen, clo, ext, idx, crit_idx, val, bnd):
        self.generator = gen
        self.closure = clo
        self.extension = ext
        self.gen_index = idx
        self.crit_idx = crit_idx
        self.val = val
        self.val_bound = bnd
        self.valid = self.crit_idx > self.gen_index

    def __repr__(self):
        return f'N({list(self.generator)}, {list(self.closure)}, {self.val:.5g}, {self.val_bound:.5g}, {list(self.extension)})'

    def value(self):
        return self.val

    def __le__(self, other):
        return self.val <= other.val

    def __eq__(self, other):
        return self.val == other.val

    def __ge__(self, other):
        return self.val >= other.val

    def __lt__(self, other):
        return self.val < other.val

    def __gt__(self, other):
        return self.val > other.val


class BreadthFirstBoundary:

    def __init__(self):
        self.deq = deque()

    def __bool__(self):
        return bool(self.deq)

    def push(self, augmented_node):
        self.deq.append(augmented_node)

    def pop(self):
        return self.deq.popleft()


class DepthFirstBoundary:

    def __init__(self):
        self.stack = []

    def __bool__(self):
        return bool(self.stack)

    def push(self, augmented_node):
        self.stack.append(augmented_node)

    def pop(self):
        return self.stack.pop()


class BestBoundFirstBoundary:

    def __init__(self):
        self.heap = []

    def __bool__(self):
        return bool(self.heap)

    def push(self, augmented_node):
        _, node = augmented_node
        heappush(self.heap, (-node.val_bound, -node.val, augmented_node))

    def pop(self):
        _, _, augmented_node = heappop(self.heap)
        return augmented_node


class BestValueFirstBoundary:

    def __init__(self):
        self.heap = []

    def __bool__(self):
        return bool(self.heap)

    def push(self, augmented_node):
        _, node = augmented_node
        heappush(self.heap, (-node.val, -node.val_bound, augmented_node))

    def pop(self):
        _, _, augmented_node = heappop(self.heap)
        return augmented_node


class Context:
    """
    Formal context, i.e., a binary relation between a set of objects and a set of attributes,
    i.e., Boolean functions that are defined on the objects.

    A formal context provides a search context (search space) over conjunctions that can be
    formed from the individual attributes.
    """

    @staticmethod
    def from_tab(table, sort_attributes=False):
        """
        Converts an input table where each row represents an object into
        a formal context (which uses column-based representation).

        Uses Boolean interpretation of table values to determine attribute
        presence for an object.

        For instance:

        >>> table = [[0, 1, 0, 1],
        ...          [1, 1, 1, 0],
        ...          [1, 0, 1, 0],
        ...          [0, 1, 0, 1]]
        >>> ctx = Context.from_tab(table)
        >>> list(ctx.extension([0, 2]))
        [1, 2]

        :param table: table that explicitly encodes object/attribute relation
        :return: context over table row indices as objects and one tabulated feature per column index
        """

        m = len(table)
        n = len(table[0])
        attributes = [TabulatedProposition(table, j) for j in range(n)]
        return Context(attributes, list(range(m)), sort_attributes)

    @staticmethod
    def from_df(df, without=None, max_col_attr=None, sort_attributes=True, discretization=pd.qcut):
        """
        Generates formal context from pandas dataframe by applying inter-ordinal scaling to numerical data columns
        and for object columns creating one attribute per value.

        For inter-ordinal scaling a maximum number of attributes per column can be specified. If required, threshold
        values are then selected by the provided discretization function (per default quantile-based).

        The restriction should also be implemented for object columns in the future (by merging small categories
        into disjunctive propositions).

        The generated attributes correspond to pandas-compatible query strings. For example:

        >>> titanic_df = pd.read_csv("../datasets/titanic/train.csv")
        >>> titanic_df.drop(columns=['PassengerId', 'Name', 'Ticket', 'Cabin'], inplace=True)
        >>> titanic_ctx = Context.from_df(titanic_df, max_col_attr=6, sort_attributes=False)
        >>> titanic_ctx.m
        891
        >>> titanic_ctx.attributes # doctest: +NORMALIZE_WHITESPACE
        [Survived<=0, Survived>=1, Pclass<=1, Pclass<=2, Pclass>=2, Pclass>=3, Sex==male, Sex==female, Age<=23.0,
        Age>=23.0, Age<=34.0, Age>=34.0, Age<=80.0, Age>=80.0, SibSp<=8.0, SibSp>=8.0, Parch<=6.0, Parch>=6.0,
        Fare<=8.6625, Fare>=8.6625, Fare<=26.0, Fare>=26.0, Fare<=512.3292, Fare>=512.3292, Embarked==S, Embarked==C,
        Embarked==Q, Embarked==nan]
        >>> titanic_ctx.n
        28
        >>> titanic_df.query('Survived>=1 & Pclass>=3 & Sex=="male" & Age>=34')
             Survived  Pclass   Sex   Age  SibSp  Parch   Fare Embarked
        338         1       3  male  45.0      0      0  8.050        S
        400         1       3  male  39.0      0      0  7.925        S
        414         1       3  male  44.0      0      0  7.925        S
        >>> titanic_ctx.extension([1, 5, 6, 11])
        array([338, 400, 414])

        (prev was SortedSet([338, 400, 414]))

        >>> titanic_ctx = Context.from_df(titanic_df, max_col_attr=defaultdict(lambda: None, Age=6, Fare=6),
        ...                               sort_attributes=False)
        >>> titanic_ctx.attributes # doctest: +NORMALIZE_WHITESPACE
        [Survived<=0, Survived>=1, Pclass<=1, Pclass<=2, Pclass>=2, Pclass>=3, Sex==male, Sex==female, Age<=23.0,
        Age>=23.0, Age<=34.0, Age>=34.0, Age<=80.0, Age>=80.0, SibSp<=0, SibSp<=1, SibSp>=1, SibSp<=2, SibSp>=2,
        SibSp<=3, SibSp>=3, SibSp<=4, SibSp>=4, SibSp<=5, SibSp>=5, SibSp>=8, Parch<=0, Parch<=1, Parch>=1, Parch<=2,
        Parch>=2, Parch<=3, Parch>=3, Parch<=4, Parch>=4, Parch<=5, Parch>=5, Parch>=6, Fare<=8.6625, Fare>=8.6625,
        Fare<=26.0, Fare>=26.0, Fare<=512.3292, Fare>=512.3292, Embarked==S, Embarked==C, Embarked==Q, Embarked==nan]


        :param df: pandas dataframe to be converted to formal context
        :param max_col_attr: maximum number of attributes generated per column;
                             or None if an arbitrary number of attributes is permitted;
                             or dict (usually defaultdict) with keys being columns ids of df and values
                             being the maximum number of attributes for the corresponding column (again using
                             None if no bound for a specific column);
                             Note: use defaultdict(lambda: None) instead of defaultdict(None) to specify no maximum
                             per default
        :param discretization: the discretization function to be used when number of thresholds has to be reduced to
                               a specificed maximum (function has to have identical signature to pandas.qcut, which
                               is the default)
        :param without: columns to ommit
        :return: context representing dataframe
        """

        without = without or []

        if not isinstance(max_col_attr, dict):
            const = max_col_attr
            max_col_attr = defaultdict(lambda: const)

        attributes = []
        for c in df:
            if c in without:
                continue
            if df[c].dtype.kind in 'uif':
                vals = df[c].unique()
                reduced = False
                max_cols = max_col_attr[str(c)]
                if max_cols and len(vals)*2 > max_cols:
                    _, vals = discretization(df[c], max_cols // 2, retbins=True, duplicates='drop')
                    vals = vals[1:]
                    reduced = True
                vals = sorted(vals)
                for i, v in enumerate(vals):
                    if reduced or i < len(vals) - 1:
                        attributes += [KeyValueProposition(c, Constraint.less_equals(v))]
                    if reduced or i > 0:
                        attributes += [KeyValueProposition(c, Constraint.greater_equals(v))]
            if df[c].dtype.kind in 'O':
                attributes += [KeyValueProposition(c, Constraint.equals(v)) for v in df[c].unique()]

        return Context(attributes, [df.iloc[i] for i in range(len(df.axes[0]))], sort_attributes)

    def __init__(self, attributes, objects, sort_attributes=True):
        self.attributes = attributes
        self.objects = objects
        self.n = len(attributes)
        self.m = len(objects)
        # for now we materialise the whole binary relation; in the future can be on demand
        # self.extents = [SortedSet([i for i in range(self.m) if attributes[j](objects[i])]) for j in range(self.n)]
        self.extents = [array([i for i in range(self.m) if attributes[j](objects[i])], dtype='int64') for j in range(self.n)]

        # sort attribute in ascending order of extent size
        if sort_attributes:
            attribute_order = list(sorted(range(self.n), key=lambda i: len(self.extents[i])))
            self.attributes = [self.attributes[i] for i in attribute_order]
            self.extents = [self.extents[i] for i in attribute_order]

    def greedy_simplification(self, intent, extent):
        to_cover = SortedSet([i for i in range(self.m) if i not in extent])
        available = list(range(len(intent)))
        covering = [SortedSet([i for i in range(self.m) if i not in self.extents[j]]) for j in intent]
        result = []
        while to_cover:
            j = max(available, key=lambda i: len(covering[i]))
            result += [intent[j]]
            available.remove(j)
            to_cover -= covering[j]
            for l in available:
                covering[l] -= covering[j]

        return result

    def extension(self, intent):
        """
        :param intent: attributes describing a set of objects
        :return: indices of objects that have all attributes in intent in common
        """
        if not intent:
            return array(range(len(self.objects)))

        result = self.extents[intent[0]]
        for i in range(1, len(intent)):
            result = snp.intersect(result, self.extents[intent[i]])

        return result

    def refinement(self, node, i, f, g, opt_val, apx=1.0):
        """
        >>> table = [[0, 1, 0, 1],
        ...          [1, 1, 1, 0],
        ...          [1, 0, 1, 0],
        ...          [0, 1, 0, 1]]
        >>> ctx = Context.from_tab(table)
        >>> f, g = lambda e: -len(e), lambda e: 1
        >>> root = Node(SortedSet([]),SortedSet([]), array([0,1,2,3]), -1, -4, 1, inf)
        >>> ref = ctx.refinement(root, 0, f, g, -4)
        >>> list(ref.closure)
        [0, 2]
        """
        # if i in node.closure:
        #     print(f"WARNING: redundant augmentation {self.attributes[i]}")
        #     return None

        generator = node.generator.copy()
        generator.add(i)
        extension = snp.intersect(node.extension, self.extents[i])

        val = f(extension)
        bound = g(extension)

        # TODO: this can apparently harm result quality: if val > opt it should still become the new
        #       opt even if the improvement (and bound) is less what is required for enqueuing
        if bound * apx < opt_val:
            return None

        closure = []
        for j in range(0, i):
            if j in node.closure:
                closure.append(j)
            elif len(extension) <= len(self.extents[j]) and \
                    len(snp.intersect(extension, self.extents[j])) == len(extension):
                return Node(generator, closure, extension, i, j, val, bound)

        closure.append(i)

        crit_idx = self.n
        for j in range(i + 1, self.n):
            if j in node.closure:
                closure.append(j)
            elif len(extension) <= len(self.extents[j]) and \
                    len(snp.intersect(extension, self.extents[j])) == len(extension):
                crit_idx = min(crit_idx, self.n)
                closure.append(j)

        return Node(generator, SortedSet(closure), extension, i, crit_idx, val, bound)

    traversal_orders = {
        'breadthfirst': BreadthFirstBoundary,
        'bestboundfirst': BestBoundFirstBoundary,
        'bestvaluefirst': BestValueFirstBoundary,
        'depthfirst': DepthFirstBoundary
    }

    def traversal(self, f, g, order='breadthfirst', apx=1.0, verbose=False):
        """
        A first example with trivial objective and bounding function is as follows. In this example
        the optimal extension is the empty extension, which is generated via the
        the lexicographically smallest and shortest generator [0, 1, 3].
        >>> table = [[0, 1, 0, 1],
        ...          [1, 1, 1, 0],
        ...          [1, 0, 1, 0],
        ...          [0, 1, 0, 1]]
        >>> ctx = Context.from_tab(table)
        >>> search = ctx.traversal(lambda e: -len(e), lambda e: 1)
        >>> for n in search:
        ...     print(n)
        N([], [], -4, inf, [0, 1, 2, 3])
        N([0], [0, 2], -2, 1, [1, 2])
        N([1], [1], -3, 1, [0, 1, 3])
        N([1, 3], [1, 3], -2, 1, [0, 3])
        N([0, 1], [0, 1, 2], -1, 1, [1])
        N([0, 1, 3], [0, 1, 2, 3], 0, 1, [])

        Let's use more realistic objective and bounding functions based on values associated with each
        object (row in the table).
        >>> values = [-1, 1, 1, -1]
        >>> f = lambda e: sum((values[i] for i in e))/4
        >>> g = lambda e: sum((values[i] for i in e if values[i]==1))/4
        >>> search = ctx.traversal(f, g)
        >>> for n in search:
        ...     print(n)
        N([], [], 0, inf, [0, 1, 2, 3])
        N([0], [0, 2], 0.5, 0.5, [1, 2])

        Finally, here is a complex example taken from the UdS seminar on subgroup discovery.
        >>> table = [[1, 1, 1, 1, 0],
        ...          [1, 1, 0, 0, 0],
        ...          [1, 0, 1, 0, 0],
        ...          [0, 1, 1, 1, 1],
        ...          [0, 0, 1, 1, 1],
        ...          [1, 1, 0, 0, 1]]
        >>> ctx = Context.from_tab(table)
        >>> labels = [1, 0, 1, 0, 0, 0]
        >>> from realkd.legacy import impact, cov_incr_mean_bound, impact_count_mean
        >>> f = impact(labels)
        >>> g = cov_incr_mean_bound(labels, impact_count_mean(labels))
        >>> search = ctx.traversal(f, g)
        >>> for n in search:
        ...     print(n)
        N([], [], 0, inf, [0, 1, 2, 3, 4, 5])
        N([0], [0], 0.11111, 0.22222, [0, 1, 2, 5])
        N([1], [1], -0.055556, 0.11111, [0, 1, 3, 5])
        N([2], [2], 0.11111, 0.22222, [0, 2, 3, 4])
        N([0, 2], [0, 2], 0.22222, 0.22222, [0, 2])

        >>> ctx.search(f, g)
        c0 & c2

        >>> labels = [1, 0, 0, 1, 1, 0]
        >>> f = impact(labels)
        >>> g = cov_incr_mean_bound(labels, impact_count_mean(labels))
        >>> search = ctx.traversal(f, g)
        >>> for n in search:
        ...     print(n)
        N([], [], 0, inf, [0, 1, 2, 3, 4, 5])
        N([0], [0], -0.16667, 0.083333, [0, 1, 2, 5])
        N([1], [1], 0, 0.16667, [0, 1, 3, 5])
        N([2], [2], 0.16667, 0.25, [0, 2, 3, 4])
        N([4], [4], 0.083333, 0.16667, [3, 4, 5])
        N([2, 3], [2, 3], 0.25, 0.25, [0, 3, 4])

        :param f: objective function
        :param g: bounding function satisfying that g(I) >= max {f(J): J >= I}
        :param order: traversal order ('breadthfirst', 'depthfirst', 'bestboundfirst' or 'bestvaluefirst')
        :param apx: approximation factor that determines guarantee of what fraction
                    of search space will be traversed, i.e., all nodes q are visited
                    with f(q) >= apx * opt (default 1.0, i.e., optimum will be visited;
                    smaller values means less traversal elements
        """
        boundary = self.traversal_orders[order]()
        full = self.extension([])
        root = Node(SortedSet([]), SortedSet([]), full, -1, self.n, f(full), inf)
        opt = root
        yield root
        boundary.push((range(self.n), root))

        k = 0
        while boundary:
            ops, current = boundary.pop()

            k += 1
            if verbose >= 2 and k % 1000 == 0:
                print('*', end='', flush=True)
            if verbose >= 1 and k % 10000 == 0:
                print(f' (best/bound: {opt.val}, {current.val_bound})', flush=True)

            children = []
            for a in ops:
                child = self.refinement(current, a, f, g, opt.val, apx)
                if child:
                    if child.valid:
                        # TODO: this is a conservative implementation that means that an
                        #       invalid child does not contribute to raising the current opt value.
                        opt = max(opt, child, key=Node.value)
                        yield child
                    children += [child]
            filtered = list(filter(lambda c: c.val_bound * apx > opt.val, children))
            ops = []
            for child in reversed(filtered):
                if child.valid:
                    boundary.push(([i for i in ops if i not in child.closure], child))
                # [i for i in ops if i not in child.closure]
                #ops = [child.gen_index] + ops
                ops = [child.gen_index] + ops

    def greedy_search(self, f, verbose=False):
        """
        >>> table = [[1, 1, 1, 1, 0],
        ...          [1, 1, 0, 0, 0],
        ...          [1, 0, 1, 0, 0],
        ...          [0, 1, 1, 1, 1],
        ...          [0, 0, 1, 1, 1],
        ...          [1, 1, 0, 0, 1]]
        >>> ctx = Context.from_tab(table)
        >>> labels = [1, 0, 1, 0, 0, 0]
        >>> from realkd.legacy import impact
        >>> f = impact(labels)
        >>> ctx.greedy_search(f)
        c0 & c2
        """
        intent = SortedSet([])
        extent = self.extension([])
        value = f(extent)
        while True:
            best_i, best_ext = None, None
            for i in range(self.n):
                if i in intent:
                    continue
                _extent = snp.intersect(extent, self.extents[i])
                _value = f(_extent)
                if _value > value:
                    value = _value
                    best_ext = _extent
                    best_i = i
            if best_i is not None:
                intent.add(best_i)
                extent = best_ext
            else:
                break
            if verbose:
                print('*', end='', flush=True)
        return Conjunction(map(lambda i: self.attributes[i], intent))

    def search(self, f, g, order='breadthfirst', apx=1.0, verbose=False):
        if verbose >= 2:
            print(f'Searching with apx factor {apx} in order {order}')
        opt = None
        opt_value = -inf
        k = 0
        for node in self.traversal(f, g, order, apx, verbose=verbose):
            k += 1
            if opt_value < node.val:
                opt = node
                opt_value = node.val
        if verbose:
            print('')
            print(f'Found optimum after inspecting {k} nodes')
        min_generator = self.greedy_simplification(opt.closure, opt.extension)
        return Conjunction(map(lambda i: self.attributes[i], min_generator))


if __name__ == '__main__':
    import doctest
    doctest.testmod()



from math import inf
from numpy import arange, cumsum
from realkd.search import Conjunction, Context, KeyValueProposition, Constraint


class Impact:
    """
    Impact objective function for conjunctive queries with respect to a specific
    dataset D and target variable y. Formally:

    impact(q) = |ext(q)|/|D| (mean(y; ext(q)) - mean(y; D)) .

    Accepts list-like, dict-like, and Pandas dataframe objects. For example:
    >>> import pandas as pd
    >>> titanic = pd.read_csv("../datasets/titanic/train.csv")
    >>> titanic.drop(columns=['PassengerId', 'Name', 'Ticket', 'Cabin'], inplace=True)
    >>> old_male = Conjunction([KeyValueProposition('Age', Constraint.greater_equals(60)),
    ...                         KeyValueProposition('Sex', Constraint.equals('male'))])
    >>> imp_survival = Impact(titanic, 'Survived')
    >>> imp_survival(old_male)
    -0.006110487591969073

    >>> imp_survival.bound(old_male)
    0.002074618236234398

    >>> imp_survival.search()
    Sex==female
    """

    def __init__(self, data, target):
        self.m = len(data)
        self.data = data.sort_values(target, ascending=False)  # data
        self.data.reset_index(drop=True, inplace=True)
        self.target = target
        self.mean = self.data[self.target].mean()

    def __call__(self, q):
        extent = self.data.loc[q]
        local_mean = extent[self.target].mean()
        return len(extent)/self.m * (local_mean - self.mean)

    def bound(self, q):
        extent = self.data.loc[q]
        data = extent[self.target]
        n = len(extent)
        if n == 0:
            return -inf
        s = cumsum(data)
        return (s - arange(1, n + 1)*self.mean).max() / self.m

    def search(self, order='breadthfirst', verbose=False):
        ctx = Context.from_df(self.data, without=[self.target], max_col_attr=10)
        return ctx.search(self, self.bound, order=order, verbose=verbose)


if __name__=='__main__':

    from timeit import timeit

    setup1 = \
"""import pandas as pd
from realkd.subgroups import Impact
titanic = pd.read_csv("../datasets/titanic/train.csv")
titanic.drop(columns=['PassengerId', 'Name', 'Ticket', 'Cabin'], inplace=True)
imp_survival = Impact(titanic, 'Survived')"""

    setup2 = \
"""import pandas as pd
from realkd.search import Impact
titanic = pd.read_csv("../datasets/titanic/train.csv")
titanic.drop(columns=['PassengerId', 'Name', 'Ticket', 'Cabin'], inplace=True)
imp_survival = Impact(titanic, 'Survived')"""

    print(timeit('imp_survival.search()', setup1, number=5))
    print(timeit('imp_survival.search()', setup2, number=5))



from pandas import Series

from sklearn.metrics import roc_auc_score as skl_auc, \
                            accuracy_score as skl_accuracy,\
                            log_loss as skl_log_loss, \
                            r2_score as skl_r2


def r2(data, target):

    def metric(model):
        return skl_r2(target, model.predict(data))

    return metric


def accuracy(data, target):

    def metric(model):
        return skl_accuracy(target, model.predict(data))

    return metric


def roc_auc(data, target):

    def metric(model):
        if callable(model):
            return skl_auc(target, model(data))
        else:
            return skl_auc(target, model.predict_proba(data)[:, 1])

    return metric


def log_loss(data, target):

    def metric(model):
        return skl_log_loss(target, model.predict_proba(data)[:, 1])

    return metric


def ensemble_length_vs_perf(model, metric):
    """ Convenience method for evaluating ensemble size / performance trade-off

    :param model: ensemble model that allows to build prefixes by slicing
    :param metric: evaluation metric function that accepts as single argument a model
    :return: pandas Series of ensemble performance for all sequences (including the empty and full prefix)
    """
    res = Series()
    for k in range(len(model.members)+1):
        _model = model[:k]
        res.at[k] = metric(_model)
    return res



import unittest
import realkd.search

from doctest import DocTestSuite


def load_tests(loader, tests, ignore):
    tests.addTests(DocTestSuite(module=realkd.search))
    return tests


if __name__ == '__main__':
    unittest.main(verbosity=2)




"""
Contains early implementation of search objective functions
and bounds mainly for speed comparison and old tests.
"""

import pandas as pd

from math import inf

from realkd.logic import Conjunction, Constraint, KeyValueProposition
from realkd.search import Context


def cov_squared_dev(labels):
    n = len(labels)
    global_mean = sum(labels) / n

    def f(count, mean):
        return count/n * pow(mean - global_mean, 2)

    return f


def impact_count_mean(labels):
    """
    >>> labels = [1, 0, 1, 0, 0, 0, 0, 0, 0, 0]
    >>> f = impact_count_mean(labels)
    >>> f(5, 0.4) # 1/2 * (2/5-1/5) = 1/6
    0.1
    """
    n = len(labels)
    m0 = sum(labels)/n

    def f(c, m):
        return c/n * (m - m0)

    return f


class DfWrapper:

    def __init__(self, df): self.df = df

    def __getitem__(self, item): return self.df.iloc[item]

    def __len__(self): return len(self.df)

    def __iter__(self):
        return (r for (_, r) in self.df.iterrows())


class Impact:
    """
    Impact objective function for conjunctive queries with respect to a specific
    dataset D and target variable y. Formally:

    impact(q) = |ext(q)|/|D| (mean(y; ext(q)) - mean(y; D)) .

    Accepts list-like, dict-like, and Pandas dataframe objects. For example:
    >>> titanic = pd.read_csv("../datasets/titanic/train.csv")
    >>> titanic.drop(columns=['PassengerId', 'Name', 'Ticket', 'Cabin'], inplace=True)
    >>> old_male = Conjunction([KeyValueProposition('Age', Constraint.greater_equals(60)),
    ...                         KeyValueProposition('Sex', Constraint.equals('male'))])
    >>> imp_survival = Impact(titanic, 'Survived')
    >>> imp_survival(old_male)
    -0.006110487591969073
    >>> imp_survival.search(verbose=True)
    <BLANKLINE>
    Found optimum after inspecting 92 nodes
    Sex==female
    """

    def _mean(self, q):
        s, c = 0.0, 0.0
        for r in filter(q, self.data):
            s += r[self.target]
            c += 1
        return s/c

    def _coverage(self, q):
        return sum(1 for _ in filter(q, self.data))/self.m

    def __init__(self, data, target):
        self.m = len(data)
        self.data = DfWrapper(data) if isinstance(data, pd.DataFrame) else data
        self.target = target
        self.average = self._mean(lambda _: True)

    def __call__(self, q):
        return self._coverage(q) * (self._mean(q) - self.average)

    def search(self, verbose=False):
        ctx = Context.from_df(self.data.df, without=[self.target], max_col_attr=10)
        f = impact(self.data.df[self.target])
        g = cov_incr_mean_bound(self.data.df[self.target], impact_count_mean(self.data.df[self.target]))
        return ctx.search(f, g, verbose=verbose)


class SquaredLossObjective:
    """
    Rule boosting objective function for squared loss.

    >>> titanic = pd.read_csv("../datasets/titanic/train.csv")
    >>> titanic.drop(columns=['PassengerId', 'Name', 'Ticket', 'Cabin'], inplace=True)
    >>> obj = SquaredLossObjective(titanic, titanic['Survived'])
    >>> female = Conjunction([KeyValueProposition('Sex', Constraint.equals('female'))])
    >>> first_class = Conjunction([KeyValueProposition('Pclass', Constraint.less_equals(1))])
    >>> obj(female)
    0.19404590848327577
    >>> reg_obj = SquaredLossObjective(titanic.drop(columns=['Survived']), titanic['Survived'], reg=2)
    >>> reg_obj(female)
    0.19342988972618597
    >>> reg_obj(first_class)
    0.09566220318908493
    >>> reg_obj._mean(female)
    0.7420382165605095
    >>> reg_obj._mean(first_class)
    0.6296296296296297
    >>> reg_obj.search()
    Sex==female
    """

    def __init__(self, data, target, reg=0):
        """
        :param data:
        :param target: _series_ of target values of matching dimension
        :param reg:
        """
        self.m = len(data)
        self.data = DfWrapper(data) if isinstance(data, pd.DataFrame) else data
        self.target = target
        self.reg = reg

    def _f(self, count, mean):
        return self._reg_term(count)*count/self.m * pow(mean, 2)

    def _reg_term(self, c):
        return 1 / (1 + self.reg / (2 * c))

    def _count(self, q): #almost code duplication: Impact
        return sum(1 for _ in filter(q, self.data))

    def _mean(self, q): #code duplication: Impact
        s, c = 0.0, 0.0
        for i in range(self.m):
            if q(self.data[i]):
                s += self.target[i]
                c += 1
        return s/c

    def search(self, max_col_attr=10):
        # here we need the function in list of row indices; can we save some of these conversions?
        def f(rows):
            c = len(rows)
            if c == 0:
                return 0.0
            m = sum(self.target[i] for i in rows) / c
            return self._f(c, m)

        g = cov_mean_bound(self.target, lambda c, m: self._f(c, m))

        ctx = Context.from_df(self.data.df, max_col_attr=max_col_attr)
        return ctx.search(f, g)

    def opt_value(self, rows):
        s, c = 0.0, 0
        for i in rows:
            s += self.target[i]
            c += 1

        return s / (self.reg/2 + c) if (c > 0 or self.reg > 0) else 0.0

    def __call__(self, q):
        c = self._count(q)
        m = self._mean(q)
        return self._f(c, m)


def impact(labels):
    """
    Compiles objective function for extension I defined by
    f(I) = len(I)/n (mean_I(l)-mean(l)) for some set of labels l of size n.

    >>> labels = [1, 0, 1, 0, 0, 0, 0, 0, 0, 0]
    >>> f = impact(labels)
    >>> f([0, 1, 2, 3, 4]) # 0.5 * (0.4 - 0.2)
    0.1
    >>> f(range(len(labels)))
    0.0
    """
    g = impact_count_mean(labels)

    def f(extension):
        if len(extension) == 0:
            return -inf
        m = sum((labels[i] for i in extension))/len(extension)
        return g(len(extension), m)

    return f


def squared_loss_obj(labels):
    """
    Builds objective function that maps index set to product
    of relative size of index set times squared difference
    of mean label value described by index set to overall
    mean label value. For instance:

    >>> labels = [-4, -2, -1, -1, 0, 1, 10, 21]
    >>> sum(labels)/len(labels)
    3.0
    >>> obj = squared_loss_obj(labels)
    >>> obj([4, 5, 6, 7])  # local avg 8, relative size 1/2
    12.5

    :param labels: y-values
    :return: f(I) = |I|/n * (mean(y)-mean_I(y))^2
    """

    f = cov_squared_dev(labels)

    def label(i): return labels[i]

    def obj(extent):
        k = len(extent)
        local_mean = sum(map(label, extent)) / k

        return f(k, local_mean)

    return obj


def cov_incr_mean_bound(labels, f):
    """
    >>> labels = [1, 1, 1, 0, 1, 0, 1, 0, 0, 0]
    >>> f = impact_count_mean(labels)
    >>> g = cov_incr_mean_bound(labels, f)
    >>> g(range(len(labels)))
    0.25
    """

    def label(i): return labels[i]

    def bound(extent):
        ordered = sorted(extent, key=label)
        k = len(ordered)
        opt = -inf

        s = 0
        for i in range(k):
            s += labels[ordered[-i-1]]
            opt = max(opt, f(i+1, s/(i+1)))

        return opt

    return bound


def cov_mean_bound(labels, f):
    """
    >>> labels = [-13, -2, -1, -1, 0, 1, 19, 21]
    >>> f = cov_squared_dev(labels)
    >>> obj = squared_loss_obj(labels)
    >>> obj(range(6,8))
    72.25
    >>> f(2, 20.0)
    72.25
    >>> bound = cov_mean_bound(labels, f)
    >>> bound(range(len(labels)))  # local avg 8, relative size 1/2
    72.25

    :param labels:
    :param f: any function that can be re-written as the maximum f(c, m)=max(g(c,m), h(c,m)) over functions g and h
              where g is monotonically increasing in its first and second argument (count and mean)
              and h is monotonically increasing in its first argument and monotonically decreasing in its second
              argument
    :return: bounding function that returns for any set of indices I, the maximum f-value over subsets J <= I
             where f is evaluated as f(|J|, mean(labels; J))
    """

    def label(i): return labels[i]

    def bound(extent):
        ordered = sorted(extent, key=label)
        k = len(ordered)
        opt = -inf

        s = 0
        for i in range(k):
            s += labels[ordered[-i-1]]
            opt = max(opt, f(i+1, s/(i+1)))

        s = 0
        for i in range(k):
            s += labels[ordered[i]]
            opt = max(opt, f(i+1, s/(i+1)))

        return opt

    return bound




"""
Contains elements of propositional logic: constraints, propositions, and
conjunctions.
"""

import pandas as pd
import re

from numpy import logical_and, ones


class Constraint:
    """
    Boolean condition on a single value with string representation. For example:
    >>> t = 21
    >>> c = Constraint.less_equals(21)
    >>> c
    Constraint(x<=21)
    >>> format(c, 'age')
    'age<=21'
    >>> c(18)
    True
    >>> c(63)
    False
    >>> import numpy as np
    >>> a =  np.arange(15, 25)
    >>> c(a)
    array([ True,  True,  True,  True,  True,  True,  True, False, False,
           False])
    """

    def __init__(self, cond, str_repr=None):
        self.cond = cond
        self.str_repr = str_repr or (lambda vn: str(cond)+'('+vn+')')

    def __call__(self, value):
        return self.cond(value)

    def __format__(self, varname):
        return self.str_repr(varname)

    def __repr__(self):
        return 'Constraint('+format(self, 'x')+')'

    @staticmethod
    def less_equals(value):
        return Constraint(lambda v: v <= value, lambda n: str(n)+'<='+str(value))

    @staticmethod
    def less(value):
        return Constraint(lambda v: v < value, lambda n: str(n)+'<'+str(value))

    @staticmethod
    def greater_equals(value):
        return Constraint(lambda v: v >= value, lambda n: str(n)+'>='+str(value))

    @staticmethod
    def greater(value):
        return Constraint(lambda v: v > value, lambda n: str(n)+'>'+str(value))

    @staticmethod
    def equals(value):
        return Constraint(lambda v: v == value, lambda n: str(n)+'=='+str(value))

    @staticmethod
    def not_equals(value):
        return Constraint(lambda v: v != value, lambda n: str(n)+'!='+str(value))


_operator_factory = {
    '==': Constraint.equals,
    '!=': Constraint.not_equals,
    '>': Constraint.greater,
    '<': Constraint.less,
    '>=': Constraint.greater_equals,
    '<=': Constraint.less_equals
}


def constraint_from_op_string(op, value):
    return _operator_factory[op](value)


class KeyValueProposition:
    """
    Callable proposition that represents constraint on value for some fixed key in a dict-like object
    such as Pandas row series.

    For example:
    >>> titanic = pd.read_csv("../datasets/titanic/train.csv")
    >>> titanic.drop(columns=['PassengerId', 'Name', 'Ticket', 'Cabin'], inplace=True)
    >>> male = KeyValueProposition('Sex', Constraint.equals('male'))
    >>> male
    Sex==male

    ---> WARNING: string values need probably be quoted in representation to work as pandas query as intended

    >>> titanic.iloc[10]
    Survived         1
    Pclass           3
    Sex         female
    Age              4
    SibSp            1
    Parch            1
    Fare          16.7
    Embarked         S
    Name: 10, dtype: object
    >>> male(titanic.iloc[10])
    False
    >>> titanic.loc[male]
         Survived  Pclass   Sex   Age  SibSp  Parch     Fare Embarked
    0           0       3  male  22.0      1      0   7.2500        S
    4           0       3  male  35.0      0      0   8.0500        S
    5           0       3  male   NaN      0      0   8.4583        Q
    6           0       1  male  54.0      0      0  51.8625        S
    7           0       3  male   2.0      3      1  21.0750        S
    ..        ...     ...   ...   ...    ...    ...      ...      ...
    883         0       2  male  28.0      0      0  10.5000        S
    884         0       3  male  25.0      0      0   7.0500        S
    886         0       2  male  27.0      0      0  13.0000        S
    889         1       1  male  26.0      0      0  30.0000        C
    890         0       3  male  32.0      0      0   7.7500        Q
    <BLANKLINE>
    [577 rows x 8 columns]

    >>> male2 = KeyValueProposition('Sex', Constraint.equals('male'))
    >>> female = KeyValueProposition('Sex', Constraint.equals('female'))
    >>> infant = KeyValueProposition('Age', Constraint.less_equals(4))
    >>> male == male2, male == infant
    (True, False)
    >>> male <= female, male >= female, infant <= female
    (False, True, True)
    """

    def __init__(self, key, constraint):
        self.key = key
        self.constraint = constraint
        self.repr = format(constraint, key)

    def __call__(self, row):
        return self.constraint(row[self.key])

    def __repr__(self):
        return self.repr

    def __eq__(self, other):
        return str(self) == str(other)

    def __le__(self, other):
        return str(self) <= str(other)


class TabulatedProposition:

    def __init__(self, table, col_idx):
        self.table = table
        self.col_idx = col_idx
        self.repr = 'c'+str(col_idx)

    def __call__(self, row_idx):
        return self.table[row_idx][self.col_idx]

    def __repr__(self):
        return self.repr


class Conjunction:
    """
    Conjunctive aggregation of propositions.

    For example:
    >>> old = KeyValueProposition('age', Constraint.greater_equals(60))
    >>> male = KeyValueProposition('sex', Constraint.equals('male'))
    >>> high_risk = Conjunction([male, old])
    >>> stephanie = {'age': 30, 'sex': 'female'}
    >>> erika = {'age': 72, 'sex': 'female'}
    >>> ron = {'age': 67, 'sex': 'male'}
    >>> high_risk(stephanie), high_risk(erika), high_risk(ron)
    (False, False, True)

    Elements can be accessed via index and are sorted lexicographically.
    >>> high_risk
    age>=60 & sex==male
    >>> high_risk[0]
    age>=60
    >>> len(high_risk)
    2

    >>> high_risk2 = Conjunction([old, male])
    >>> high_risk == high_risk2
    True

    >>> titanic = pd.read_csv("../datasets/titanic/train.csv")
    >>> titanic.drop(columns=['PassengerId', 'Name', 'Ticket', 'Cabin'], inplace=True)
    >>> male = KeyValueProposition('Sex', Constraint.equals('male'))
    >>> third_class = KeyValueProposition('Pclass', Constraint.greater_equals(3))
    >>> conj = Conjunction([male, third_class])
    >>> titanic.loc[conj]
         Survived  Pclass   Sex   Age  SibSp  Parch     Fare Embarked
    0           0       3  male  22.0      1      0   7.2500        S
    4           0       3  male  35.0      0      0   8.0500        S
    5           0       3  male   NaN      0      0   8.4583        Q
    7           0       3  male   2.0      3      1  21.0750        S
    12          0       3  male  20.0      0      0   8.0500        S
    ..        ...     ...   ...   ...    ...    ...      ...      ...
    877         0       3  male  19.0      0      0   7.8958        S
    878         0       3  male   NaN      0      0   7.8958        S
    881         0       3  male  33.0      0      0   7.8958        S
    884         0       3  male  25.0      0      0   7.0500        S
    890         0       3  male  32.0      0      0   7.7500        Q
    <BLANKLINE>
    [347 rows x 8 columns]
    """

    def __init__(self, props):
        self.props = sorted(props, key=str)
        self.repr = str.join(" & ", map(str, self.props)) if props else 'True'

    def __call__(self, x):
        if not self.props:
            return ones(len(x), dtype='bool')  # TODO: check if this is correct handling for scalar x
        return logical_and.reduce([p(x) for p in self.props])
        # res = ones(len(x), dtype='bool')
        # for p in self.props:
        #     res &= p(x)
        # return res
        #return all(map(lambda p: p(x), self.props))

    def __repr__(self):
        return self.repr

    def __getitem__(self, item):
        return self.props[item]

    def __len__(self):
        return len(self.props)

    def __eq__(self, other):
        """
        Checks equality of conjunctions based on string representation.
        """
        return str(self) == str(other)


if __name__ == '__main__':
    import doctest
    doctest.testmod()



import numpy as np

from rulefit import RuleFit, Rule
from matplotlib.patches import Rectangle

from realkd.logic import Conjunction, KeyValueProposition, constraint_from_op_string

operator_bound = {
    '<=': 1,  # upper bound
    '<': 1,
    '>=': 0,
    '>': 0
}

operator_comp = {
    '<=': min,  # upper bound
    '<': min,
    '>=': max,
    '>': max
}


def patch_from_rule(rule, fc='blue', axes={'x1': 0, 'x2': 1}, x_min=-4, x_max=4, y_min=-4, y_max=4):
    bounds = [[x_min, x_max], [y_min, y_max]]

    for c in rule.conditions:
        axis = axes[c.feature_name]
        aggr = operator_comp[c.operator]
        bnd = operator_bound[c.operator]
        bounds[axis][bnd] = aggr(c.threshold, bounds[axis][bnd])

    (x, y) = bounds[0][0], bounds[1][0]
    dx = bounds[0][1] - bounds[0][0]
    dy = bounds[1][1] - bounds[1][0]

    return Rectangle((x, y), dx, dy, fill=True, color='black', lw=1, ls='-', fc=fc, alpha=0.2)


def rf_rule_as_query(self):
    props = []
    for cond in self.conditions:
        constraint = constraint_from_op_string(cond.operator, cond.threshold)
        props.append(KeyValueProposition(cond.feature_name, constraint))
    return Conjunction(props)


Rule.as_realkd_query = rf_rule_as_query


def rf_predict_proba(self, X):
    """Predict outcome probability for X, if model type supports probability prediction method
    """

    if 'predict_proba' not in dir(self.lscv):

        error_message = '''
        Probability prediction using predict_proba not available for
        model type {lscv}
        '''.format(lscv=self.lscv)
        raise ValueError(error_message)

    X_concat=np.zeros([X.shape[0],0])
    if 'l' in self.model_type:
        if self.lin_standardise:
            X_concat = np.concatenate((X_concat,self.friedscale.scale(X)), axis=1)
        else:
            X_concat = np.concatenate((X_concat,X), axis=1)
    if 'r' in self.model_type:
        rule_coefs=self.coef_[-len(self.rule_ensemble.rules):]
        if len(rule_coefs)>0:
            X_rules = self.rule_ensemble.transform(X,coefs=rule_coefs)
            if X_rules.shape[0] >0:
                X_concat = np.concatenate((X_concat, X_rules), axis=1)
    return self.lscv.predict_proba(X_concat)


RuleFit.predict_proba = rf_predict_proba





import collections.abc

from math import inf
from numpy import arange, argsort, array, cumsum, exp, full_like, log2, stack, zeros, zeros_like
from pandas import qcut, Series

from realkd.search import Conjunction, Context, KeyValueProposition, Constraint


class SquaredLoss:
    """
    >>> squared_loss
    squared_loss
    >>> y = array([-2, 0, 3])
    >>> s = array([0, 1, 2])
    >>> squared_loss(s, y)
    array([4, 1, 1])
    >>> squared_loss.g(s, y)
    array([-4, -2,  2])
    >>> squared_loss.h(s, y)
    array([2, 2, 2])
    """

    _instance = None

    def __new__(cls):
        if cls._instance is None:
            cls._instance = super(SquaredLoss, cls).__new__(cls)
        return cls._instance

    @staticmethod
    def __call__(y, s):
        return (y - s)**2

    @staticmethod
    def predictions(s):
        return s

    @staticmethod
    def g(y, s):
        return -2*(y - s)

    @staticmethod
    def h(y, s):
        return full_like(s, 2)  # Series(full_like(s, 2))

    @staticmethod
    def __repr__():
        return 'squared_loss'

    @staticmethod
    def __str__():
        return 'squared'


class LogisticLoss:
    """
    Logistic loss function l(y, s) = log2(1 + exp(-ys)).

    Function assumes that positive and negative values are encoded as +1 and -1, respectively.

    >>> y = array([1, -1, 1, -1])
    >>> s = array([0, 0, 10, 10])
    >>> logistic_loss(y, s)
    array([1.00000000e+00, 1.00000000e+00, 6.54967668e-05, 1.44270159e+01])
    >>> logistic_loss.g(y, s)
    array([-5.00000000e-01,  5.00000000e-01, -4.53978687e-05,  9.99954602e-01])
    >>> logistic_loss.h(y, s)
    array([2.50000000e-01, 2.50000000e-01, 4.53958077e-05, 4.53958077e-05])
    """

    _instance = None

    def __new__(cls):
        if cls._instance is None:
            cls._instance = super(LogisticLoss, cls).__new__(cls)
        return cls._instance

    @staticmethod
    def __call__(y, s):
        return log2(1 + exp(-y*s))

    @staticmethod
    def sigmoid(a):
        return 1 / (1 + exp(-a))

    @staticmethod
    def predictions(s):
        preds = zeros_like(s)
        preds[s >= 0] = 1
        preds[s < 0] = -1
        return preds  # this case now returns np array

    @staticmethod
    def probabilities(s):
        pos = LogisticLoss.sigmoid(s)
        return stack((1-pos, pos), axis=1)

    @staticmethod
    def g(y, s):
        return -y*LogisticLoss.sigmoid(-y*s)

    @staticmethod
    def h(y, s):
        sig = LogisticLoss.sigmoid(-y*s)
        return sig*(1.0-sig)

    @staticmethod
    def __repr__():
        return 'logistic_loss'

    @staticmethod
    def __str__():
        return 'logistic'


logistic_loss = LogisticLoss()
squared_loss = SquaredLoss()


loss_functions = {
    LogisticLoss.__repr__(): logistic_loss,
    SquaredLoss.__repr__(): squared_loss,
    LogisticLoss.__str__(): logistic_loss,
    SquaredLoss.__str__(): squared_loss
}


def loss_function(loss):
    if callable(loss):
        return loss
    else:
        return loss_functions[loss]


class GradientBoostingObjective:
    """
    >>> import pandas as pd
    >>> titanic = pd.read_csv("../datasets/titanic/train.csv")
    >>> survived = titanic['Survived']
    >>> titanic.drop(columns=['PassengerId', 'Name', 'Ticket', 'Cabin', 'Survived'], inplace=True)
    >>> obj = GradientBoostingObjective(titanic, survived, reg=0.0)
    >>> female = Conjunction([KeyValueProposition('Sex', Constraint.equals('female'))])
    >>> first_class = Conjunction([KeyValueProposition('Pclass', Constraint.less_equals(1))])
    >>> obj(obj.data[female].index)
    0.1940459084832758
    >>> obj(obj.data[first_class].index)
    0.09610508375940474
    >>> obj.bound(obj.data[first_class].index)
    0.1526374859708193

    >>> reg_obj = GradientBoostingObjective(titanic, survived, reg=2)
    >>> reg_obj(reg_obj.data[female].index)
    0.19342988972618602
    >>> reg_obj(reg_obj.data[first_class].index)
    0.09566220318908492

    >>> q = reg_obj.search(verbose=True)
    <BLANKLINE>
    Found optimum after inspecting 100 nodes
    >>> q
    Sex==female
    >>> reg_obj.opt_weight(q)
    0.7396825396825397

    >>> obj = GradientBoostingObjective(titanic, survived.replace(0, -1), loss='logistic')
    >>> obj(obj.data[female].index)
    0.04077109318199465
    >>> obj.opt_weight(female)
    0.9559748427672956
    >>> best = obj.search(order='bestvaluefirst', verbose=True)
    <BLANKLINE>
    Found optimum after inspecting 443 nodes
    >>> best
    Pclass>=2 & Sex==male
    >>> obj(obj.data[best].index)
    0.13072995752734315
    >>> obj.opt_weight(best)
    -1.4248366013071896
    """

    def __init__(self, data, target, predictions=None, loss=SquaredLoss, reg=1.0):
        self.loss = loss_function(loss)
        self.reg = reg
        predictions = zeros_like(target) if predictions is None else predictions
        g = array(self.loss.g(target, predictions))
        h = array(self.loss.h(target, predictions))
        r = g / h
        order = argsort(r)[::-1]
        self.g = g[order]
        self.h = h[order]
        self.data = data.iloc[order].reset_index(drop=True)
        self.target = target.iloc[order].reset_index(drop=True)
        self.n = len(target)

    def __call__(self, ext):
        if len(ext) == 0:
            return -inf
        g_q = self.g[ext]
        h_q = self.h[ext]
        return g_q.sum() ** 2 / (2 * self.n * (self.reg + h_q.sum()))

    def bound(self, ext):
        m = len(ext)
        if m == 0:
            return -inf

        g_q = self.g[ext]
        h_q = self.h[ext]

        num_pre = cumsum(g_q)**2
        num_suf = cumsum(g_q[::-1])**2
        den_pre = cumsum(h_q) + self.reg
        den_suf = cumsum(h_q[::-1]) + self.reg
        neg_bound = (num_suf / den_suf).max() / (2 * self.n)
        pos_bound = (num_pre / den_pre).max() / (2 * self.n)
        return max(neg_bound, pos_bound)

    def opt_weight(self, q):
        # ext = self.ext(q)
        ext = self.data.loc[q].index
        g_q = self.g[ext]
        h_q = self.h[ext]
        return -g_q.sum() / (self.reg + h_q.sum())

    def search(self, order='bestboundfirst', max_col_attr=10, discretization=qcut, apx=1.0, verbose=False):
        ctx = Context.from_df(self.data, max_col_attr=max_col_attr, discretization=discretization)
        if verbose >= 2:
            print(f'Created search context with {len(ctx.attributes)} attributes')  #:\n {ctx.attributes}')
        if order == 'greedy':
            return ctx.greedy_search(self, verbose=verbose)
        else:
            return ctx.search(self, self.bound, order=order, apx=apx, verbose=verbose)


class Rule:
    """
    Represents a rule of the form "r(x) = y if q(x) else z"
    for some binary query function q.

    >>> import pandas as pd
    >>> titanic = pd.read_csv('../datasets/titanic/train.csv')
    >>> titanic[['Name', 'Sex', 'Survived']].iloc[0]
    Name        Braund, Mr. Owen Harris
    Sex                            male
    Survived                          0
    Name: 0, dtype: object
    >>> titanic[['Name', 'Sex', 'Survived']].iloc[1]
    Name        Cumings, Mrs. John Bradley (Florence Briggs Th...
    Sex                                                    female
    Survived                                                    1
    Name: 1, dtype: object

    >>> female = KeyValueProposition('Sex', Constraint.equals('female'))
    >>> r = Rule(female, 1.0, 0.0)
    >>> r(titanic.iloc[0]), r(titanic.iloc[1])
    (0.0, 1.0)
    >>> target = titanic.Survived
    >>> titanic.drop(columns=['PassengerId', 'Name', 'Ticket', 'Cabin', 'Survived'], inplace=True)
    >>> opt = Rule(reg=0.0)
    >>> opt.fit(titanic, target)
       +0.7420 if Sex==female

    >>> best_logistic = Rule(loss='logistic')
    >>> best_logistic.fit(titanic, target.replace(0, -1))
       -1.4248 if Pclass>=2 & Sex==male

    >>> best_logistic.predict(titanic) # doctest: +ELLIPSIS
    array([-1.,  1.,  1.,  1., ...,  1.,  1., -1.])

    >>> greedy = Rule(loss='logistic', reg=1.0, method='greedy')
    >>> greedy.fit(titanic, target.replace(0, -1))
       -1.4248 if Pclass>=2 & Sex==male

    >>> empty = Rule()
    >>> empty
       +0.0000 if True
    """

    # max_col attribute to change number of propositions
    def __init__(self, q=Conjunction([]), y=0.0, z=0.0, loss=SquaredLoss, reg=1.0, max_col_attr=10,
                 discretization=qcut, method='bestboundfirst', apx=1.0):
        self.q = q
        self.y = y
        self.z = z
        self.reg = reg
        self.max_col_attr = max_col_attr
        self.discretization = discretization
        self.loss = loss
        self.method = method
        self.apx = apx

    def __call__(self, x):
        sat = self.q(x)
        return sat*self.y + (1-sat)*self.z

    def __repr__(self):
        # TODO: if existing also print else part
        return f'{self.y:+10.4f} if {self.q}'

    def fit(self, data, target, scores=None, verbose=False):
        """
        Fits rule to provide best loss reduction on given data
        (where the baseline prediction scores are either given
        explicitly through the scores parameter or are assumed
        to be 0.

        :param data: pandas DataFrame containing only the feature columns
        :param target: pandas Series containing the target values
        :param scores: prior prediction scores according to which the reduction in prediction loss is optimised
        :param verbose: whether to print status update and summary of query search
        :return: self

        """
        obj = GradientBoostingObjective(data, target, predictions=scores, loss=self.loss, reg=self.reg)

        # create residuals within init. modify implementation for that
        self.q = obj.search(order=self.method, max_col_attr=self.max_col_attr, discretization=self.discretization,
                            apx=self.apx, verbose=verbose)
        self.y = obj.opt_weight(self.q)
        return self

    def predict(self, data):
        loss = loss_function(self.loss)
        return loss.predictions(self(data))

    def predict_proba(self, data):
        loss = loss_function(self.loss)
        return loss.probabilities(self(data))


class GradientBoostingRuleEnsemble:
    """
    >>> female = KeyValueProposition('Sex', Constraint.equals('female'))
    >>> r1 = Rule(Conjunction([]), -0.5, 0.0)
    >>> r2 = Rule(female, 1.0, 0.0)
    >>> r3 = Rule(female, 0.3, 0.0)
    >>> r4 = Rule(Conjunction([]), -0.2, 0.0)
    >>> ensemble = GradientBoostingRuleEnsemble(members=[r1, r2, r3, r4])
    >>> len(ensemble)
    4
    >>> ensemble[2]
       +0.3000 if Sex==female

    >>> ensemble[:2]
       -0.5000 if True
       +1.0000 if Sex==female

    >>> import pandas as pd
    >>> titanic = pd.read_csv('../datasets/titanic/train.csv')
    >>> survived = titanic.Survived
    >>> titanic.drop(columns=['PassengerId', 'Name', 'Ticket', 'Cabin', 'Survived'], inplace=True)
    >>> re = GradientBoostingRuleEnsemble(loss=logistic_loss)
    >>> re.fit(titanic, survived.replace(0, -1), verbose=0) # doctest: +SKIP
       -1.4248 if Pclass>=2 & Sex==male
       +1.7471 if Pclass<=2 & Sex==female
       +2.5598 if Age<=19.0 & Fare>=7.8542 & Parch>=1.0 & Sex==male & SibSp<=1.0

    # performance with bestboundfirst:
    # <BLANKLINE>
    # Found optimum after inspecting 443 nodes
    #    -1.4248 if Pclass>=2 & Sex==male
    # <BLANKLINE>
    # Found optimum after inspecting 786 nodes
    #    +1.7471 if Pclass<=2 & Sex==female
    # ******
    # Found optimum after inspecting 6564 nodes
    #

    >>> re_with_offset = GradientBoostingRuleEnsemble(max_rules=2, loss='logistic', offset_rule=True)
    >>> re_with_offset.fit(titanic, survived.replace(0, -1))
       -0.4626 if True
       +2.3076 if Pclass<=2 & Sex==female

    >>> greedy = GradientBoostingRuleEnsemble(max_rules=3, loss='logistic', method='greedy')
    >>> greedy.fit(titanic, survived.replace(0, -1)) # doctest: +SKIP
       -1.4248 if Pclass>=2 & Sex==male
       +1.7471 if Pclass<=2 & Sex==female
       -0.4225 if Parch<=1.0 & Sex==male
    """

    def __init__(self, max_rules=3, loss=SquaredLoss, members=[], reg=1.0, max_col_attr=10, discretization=qcut,
                 offset_rule=False, method='bestboundfirst', apx=1.0):
        self.reg = reg
        self.members = members[:]
        self.max_col_attr = max_col_attr
        self.max_rules = max_rules
        self.discretization = discretization
        self.loss = loss
        self.offset_rule = offset_rule
        self.method = method
        if callable(apx):
            self.apx = apx
        elif isinstance(apx, collections.abc.Sequence):
            self.apx = lambda i: apx[min(i, len(apx)-1)]
        else:
            self.apx = lambda _: apx

    def __call__(self, x):  # look into swapping to Series and numpy
        res = zeros(len(x))  # TODO: a simple reduce should do if we can rule out empty ensemble
        for r in self.members:
            res += r(x)
        return res
        #return sum(r(x) for r in self.members)

    def __repr__(self):
        return str.join('\n', (str(r) for r in self.members))

    def __len__(self):
        return len(self.members)

    def __getitem__(self, item):
        """ Index access to the individual members of the ensemble.

        :param item: index
        :return: rule of index
        """
        if isinstance(item, slice):
            _members = self.members[item]
            return GradientBoostingRuleEnsemble(self.max_rules, self.loss, _members, self.reg, self.max_col_attr,
                                                self.discretization, self.offset_rule, self.method, self.apx)
        else:
            return self.members[item]

    def fit(self, data, target, verbose=False):
        if len(self.members) < self.max_rules and self.offset_rule:
            obj = GradientBoostingObjective(data, target, loss=self.loss, reg=self.reg)
            q = Conjunction([])
            y = obj.opt_weight(q)
            r = Rule(q=q, y=y, loss=self.loss, reg=self.reg, method=self.method)
            if verbose:
                print(r)
            self.members.append(r)

        while len(self.members) < self.max_rules:
            scores = self(data)
            apx = self.apx(len(self.members))
            r = Rule(loss=self.loss, reg=self.reg, max_col_attr=self.max_col_attr, discretization=self.discretization,
                     method=self.method, apx=apx)
            r.fit(data, target, scores, verbose)
            if verbose:
                print(r)
            self.members.append(r)
        return self

    def predict(self, data):
        loss = loss_function(self.loss)
        return loss.predictions(self(data))

    def predict_proba(self, data):
        loss = loss_function(self.loss)
        return loss.probabilities(self(data))

    def size(self):
        return sum(len(r.q) for r in self.members) + len(self.members)

    def consolidated(self, inplace=False):
        """ Consolidates rules with equivalent queries into one.

        :param inplace: whether to update self or to create new ensemble
        :return: reference to consolidated ensemble (self if inplace=True)

        For example:
        >>> female = KeyValueProposition('Sex', Constraint.equals('female'))
        >>> r1 = Rule(Conjunction([]), -0.5, 0.0)
        >>> r2 = Rule(female, 1.0, 0.0)
        >>> r3 = Rule(female, 0.3, 0.0)
        >>> r4 = Rule(Conjunction([]), -0.2, 0.0)
        >>> ensemble = GradientBoostingRuleEnsemble(members=[r1, r2, r3, r4])
        >>> ensemble.consolidated(inplace=True) # doctest: +NORMALIZE_WHITESPACE
        -0.7000 if True
        +1.3000 if Sex==female
        """

        _members = self.members[:]
        for i, r1 in enumerate(_members):
            q = r1.q
            y = r1.y
            z = r1.z
            for j in range(len(_members)-1, i, -1):
                r2 = _members[j]
                if q == r2.q:
                    y += r2.y
                    z += r2.z
                    _members.pop(j)
            _members[i] = Rule(q, y, z)

        if inplace:
            self.members = _members
            return self
        else:
            return GradientBoostingRuleEnsemble(self.max_rules, self.loss, _members, self.reg, self.max_col_attr,
                                                self.discretization, self.offset_rule, self.method, self.apx)


if __name__ == '__main__':

    from timeit import timeit

    setup1 = \
"""import pandas as pd
from realkd.rules import GradientBoostingObjective
titanic = pd.read_csv("../datasets/titanic/train.csv")
sql_survival = GradientBoostingObjective(titanic.drop(columns=['PassengerId', 'Name', 'Ticket', 'Cabin', 'Survived']), titanic['Survived'], reg=2)"""

    setup2 = \
"""import pandas as pd
from realkd.legacy import SquaredLossObjective
titanic = pd.read_csv("../datasets/titanic/train.csv")
sql_survival = SquaredLossObjective(titanic.drop(columns=['PassengerId', 'Name', 'Ticket', 'Cabin', 'Survived']), titanic['Survived'], reg=2)"""

    print(timeit('sql_survival.search()', setup1, number=5))
    print(timeit('sql_survival.search()', setup2, number=5))



import numpy as np
import pandas as pd


def noisy_parity(n, d=3, variance=0.25, as_df=True, random_seed=None):
    """Generates observations of mixture model of Gaussian clusters centered
    at nodes of hypercube {-1, 1}^d labelled according to parity of cube node.

    For example:
    >>> x, y = noisy_parity(10, random_seed=0)
    >>> x
             x1        x2        x3
    0  0.633866  0.727871  0.841850
    1 -0.794185 -0.478743 -1.064267
    2 -0.316768 -1.332597 -0.824245
    3  1.451735  1.047006  0.628250
    4  0.539137  0.771137  1.110098
    5  0.495191  0.895412  0.920387
    6  1.270423  1.107330 -0.822314
    7  0.673086  0.935193 -0.608012
    8 -0.253284  0.370467  1.756962
    9 -0.327062  1.390656  1.132228
    >>> y
    0    1
    1   -1
    2   -1
    3    1
    4    1
    5    1
    6   -1
    7   -1
    8   -1
    9   -1
    dtype: int64

    :param n: number of observations
    :param d: dimension of data
    :param variance: variance of the clusters
    :param as_df: whether to wrap return value in pandas dataframe/series
    :param random_seed:
    :return: dataframe/matrix x and corresponding label series/arrays
    """
    rng = np.random.default_rng(seed=random_seed)
    c = rng.integers(0, 1, (n, d), endpoint=True)
    c[c == 0] = -1
    y = np.multiply.reduce(c.T)
    x = np.apply_along_axis(rng.multivariate_normal, 1, c, variance*np.eye(d))
    if as_df:
        return pd.DataFrame(x, columns=[f'x{i+1}' for i in range(d)]), pd.Series(y)
    else:
        return x, y




from setuptools import setup

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='realkd',
    version='0.1.0',
    packages=['realkd'],
    url='https://github.com/marioboley/realkd.py',
    license='MIT',
    author='Mario Boley',
    author_email='mario.boley@gmail.com',
    description='Python implementation of knowledge discovery methods',
    long_description=long_description,
    long_description_content_type="text/markdown",
    python_requires='>=3.6',
    install_requires=['sortedcontainers>=2.1.0',
                      'pandas>=0.25',
                      'numpy>=1.16.1',
                      'matplotlib',
                      'sortednp>=0.3.0'],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3'
    ]
)
